//=========================================
// File name: quizfaber.js
//-----------------------------------------
// Project : QuizFaber 
// Licence : MIT
// Author  : Luca Galli
// Email   : info@quizfaber.com
//=========================================

var fileSystem = require('fs');
var express = require('express');
var app = express();
var bodyParser = require('body-parser');
var urlencodedParser = bodyParser.urlencoded({ extended: true });

var port = process.env.PORT || 3313;
var useSession = false;

const dbConfigFile = "db.conf";
const homePageFile = "home.html";
const anonymousLogin = "anonymous@anonymous.nowhere";

const { Login, Quiz, Result, ResultHeader, ResultItem, Abstract, SearchParams, RetakeInfo, SessionInfo } = require('./definitions');
const { GetQueryStringForCheckResult, GetQueryStringSelectPersor, GetQueryStringInsertPerson, GetQueryStringFromQuiz, GetQueryStringForInsertQuizResultReport, GetQueryStringForQuizResultReport, GetQueryStringForAllAnswers, GetQueryStringFromQuestion, GetQueryStringForQuizReport, GetQueryStringForQuestionReport, GetQueryStringForGetTitles, GetQueryStringForGetResultsAbstract, GetQueryStringForUpdateDateLastAccess, GetQueryStringForGetDbVersion, GetQueryStringForDbSize, GetQueryStringOpenSession, GetQueryStringUpdateDataSession, GetQueryStringCloseSession, GetQueryStringGetLastOpenedSession } = require('./queries');

const mariadb = require('mariadb');
var pool;
var isDbConnected = false;

var qfDbVersion = 0;
var qfDbSize = 0;

var pjson = require('./package.json');

/// <summary>
/// Running server details.
/// </summary>
var server = app.listen(port, function ()
{
	var host = server.address().address;

	console.log("###############################################");
	console.log("#     QuizFaber project - Node.js server      #");
	console.log("###############################################");
	console.log("");
	console.log("Listening at host %s port %s", host, port);
	console.log("Node.js version : " + process.version);
	console.log("Package version : " + pjson.version);

	// read db config file and create a pool
	fileSystem.readFile(dbConfigFile, 'utf8', function (err, data)
	{
		if (err) {
			console.log('db config file not found!');
			console.error(err);
			throw err;
		}
		
		console.log('found db config file');

		var dbConfigData = JSON.parse(data);
		
		InitMariaDB(dbConfigData);
	});	
});

/// <summary>
/// Database inizialization 
/// </summary>
function InitMariaDB(dbConfigData)
{
	console.log("connecting to db host : " + dbConfigData.dbHost + "..");

	// create connection pool
	pool = mariadb.createPool({
		host: dbConfigData.dbHost,
		user: dbConfigData.dbUser,
		password: dbConfigData.dbPassword,
		database: dbConfigData.dbName,
		connectionLimit: 5,
		multipleStatements: true
		//timezone: 'UTC'
	});

	useSession = dbConfigData.useSession;

	console.log("created connection pool for db : " + dbConfigData.dbName);

	// set the time zone
	pool.on('connection', conn => {
		conn.query("SET time_zone='+00:00';", err => {
			if (err) {
				console.log('cannot set timezone');
				console.error('ERROR : ' + err);
				throw err;
			}
		})
	});

	// check connection , read table 'Settings' and check db size
	pool.getConnection()
		.then(conn => {
			isDbConnected = true;
			conn.query(GetQueryStringForGetDbVersion())
				.then((rows) => {
					if (rows.length >= 1) {
						qfDbVersion = rows[0].DbVersion;
						console.log("found QuizFaber DB version : " + qfDbVersion);

						conn.query(GetQueryStringForDbSize(dbConfigData.dbName))
							.then((rows) => {
								
								if (rows.length >= 1) {
									qfDbSize = rows[0].dbsize;
									console.log("found QuizFaber DB size : " + qfDbSize);
								}
								else {
									console.error("ERROR : unable to check QuizFaber DB size");
								}
								conn.end();
							})
							.catch(err => {
								// select db size failed
								console.log('cannot check database size');
								console.error('ERROR : ' + err);
								conn.end();
							})
					}
					else {
						console.error("ERROR : unable to find QuizFaber DB version");
						conn.end();
					}
				})
				.catch(err => {
					// select db version failed
					console.log('cannot select from table Settings');
					console.error('ERROR : ' + err);
					conn.end();
				})
		})
		.catch(err => {
			//not connected
			console.error('cannot connect to database : ' + err);
			throw err;
		});
}

/// <summary>
/// Bypass the CORS policy: No 'Access-Control-Allow-Origin' header is present on the requested resource.
/// </summary>
app.use(function (req, res, next)
{
	res.header("Access-Control-Allow-Origin", "*"); // update to match the domain you will make the request from
	res.header("Access-Control-Allow-Headers", "Origin, X-Requested-With, Content-Type, Accept");
	next();
});

/// <summary>
/// Getting the home page
/// </summary>
app.get('/', function (request, response) {
	console.log("GET index");
	SendHomeResponse(response);
});

/// <summary>
/// Returns the html of the home page
/// </summary>
function SendHomeResponse(res) {

	// read home page file
	fileSystem.readFile(homePageFile, 'utf8', function (err, data) {
		if (err) {
			console.log('WARNING : home page file not found!');
			var html = "<body><h1>QuizFaber Node.js server up and running</h1></body>";
			res.send(html);
		}
		else {
			console.log('found home page file');

			if (isDbConnected) {
				if (qfDbVersion > 0) {
					data = data.replace("{ERR_CONFIG}", "");
				}
				else {
					data = data.replace("{ERR_CONFIG}", "<div class='alert alert-warning'><b>Database is not configurated</b>. Popolate the database running the sql script <b>DB_QF_CreateTables.MySQL.sql</b></div>");
				}
			}
			else {
				data = data.replace("{ERR_CONFIG}", "<div class='alert alert-danger'><b>Not connected to database</b>. Check database connection parameters into <b>db.conf</b> file</div>");
			}

			data = data.replace("{NODE_VERS}", process.version);
			data = data.replace("{PKG_VERS}", pjson.version);

			if (qfDbVersion > 0) {
				data = data.replace("{DB_VERS}", qfDbVersion);
			}
			else {
				data = data.replace("{DB_VERS}", "<span style='color:red'>Unknown</span>");
			}

			if (qfDbSize > 0) {
				data = data.replace("{DB_SIZE}", qfDbSize + " MBytes");
			}
			else {
				data = data.replace("{DB_SIZE}", "<span style='color:red'>Unknown</span>");
			}

			res.writeHead(200, { 'Content-Type': 'text/html' });
			res.write(data);
			res.end();
		}
	});
}

/// <summary>
/// Store quiz results sent by QuizFaber client
/// </summary>
app.post('/results', urlencodedParser, function (req, res)
{
	console.log('POST Results');

	try
	{
		var quiz = JSON.parse(req.body.quiz);

		pool.getConnection()
			.then(conn => {

				conn.query("START TRANSACTION")
					.then(() => {

						var query = GetQueryStringFromQuiz(quiz);

						conn.query(query)
							.then( () => {

								// get the assigned quiz id from autoincrement field
								var queryQuizID = "SELECT LAST_INSERT_ID() AS QUIZID";
								conn.query(queryQuizID)
									.then((rows) => {

										var quizID = rows[0].QUIZID;
										console.log('insert quiz row done, quiz id = ' + quizID);
										console.log('num of questions to store =' + quiz.questions.length);

										var numQuestDone = 0;

										for (var i = 0; i < quiz.questions.length; i++) {
											var question = quiz.questions[i];

											query = GetQueryStringFromQuestion(quizID, i + 1, question);

											conn.query(query + "; SELECT ID,QuestNum FROM QuizResultQuestion WHERE ID = LAST_INSERT_ID()")
												.then((rows) => {

													var questionID = rows[1][0].ID;
													var questNum = rows[1][0].QuestNum;
													var question = quiz.questions[questNum - 1];

													console.log('insert question row done, id=' + questionID + ", num=" + questNum);

													query = GetQueryStringForAllAnswers(question, questionID);

													conn.query(query)
														.then(async() => {
															console.log('insert answers of question done');

															numQuestDone++;
															if (numQuestDone === quiz.questions.length) {
																// all done
																
																console.log('all answers row inserted');

																console.log('insert quiz report');

																var queryReport = GetQueryStringForInsertQuizResultReport(quizID, req.body.report);
																await conn.query(queryReport);

																if ((quiz.currentUser !== null) && (useSession))
																{
																	if (quiz.currentUser.email !== anonymousLogin)
																	{
																		console.log('get id for user ' + quiz.currentUser.email);
																		var queryUser = GetQueryStringSelectPersor(quiz.currentUser.email);
																		var userResult = await conn.query(queryUser);
																		var personID = userResult[0].ID;

																		console.log('update session for ID=' + personID);
																		var querySession = GetQueryStringCloseSession(personID, quiz.currentUser.sessionId, GetRemoteIp(req), quizID)
																		await conn.query(querySession);
																	}
																}

																conn.query("COMMIT").then(() => { conn.end(); });
																console.log('all done');
																res.send('OK');
															}
														})
														.catch(err => {
															HandleErrorInTransaction(res, conn, 'error in inserting answer', err);
														});
												})
												.catch(err => {
													HandleErrorInTransaction(res, conn, 'error in inserting question', err);
												});
										}
									})
									.catch(err => {
										HandleErrorInTransaction(res, conn, 'error in get quiz result id', err);
									});
							})
							.catch(err => {
								HandleErrorInTransaction(res, conn, 'error in inserting quiz row', err);
							});

					}).catch(err => {
						// start transaction failed
						console.log('transaction not started : ' + err);
						conn.end();
						SendInternalServerError(res);
					});

			}).catch(err => {
				//not connected
				console.log('not connected : ' + err);
				SendInternalServerError(res);
			});
	}
	catch (err)
	{
		console.log('exception : ' + err);
		SendInternalServerError(res);
	}
});

/// <summary>
/// New user registration   
/// </summary>
app.post('/registration', urlencodedParser, function (req, res) {

	console.log('User registration');

	try
	{
		var user = JSON.parse(req.body.login);

		pool.getConnection()
			.then(conn => {

				var query = GetQueryStringInsertPerson(user);
				conn.query(query)
					.then(() => {
						console.log('insert user done');
						conn.end(); 
						res.send('OK');
					})
					.catch(err => {
						HandleError(res, conn, 'error in inserting user', err);
					});
			}).catch(err => {
				//not connected
				console.log('not connected : ' + err);
				SendInternalServerError(res);
			});
	}
	catch (err)
	{
		console.log('exception : ' + err);
		SendInternalServerError(res);
	}
});

/// <summary>
/// User login   
/// </summary>
app.post('/login', urlencodedParser, function (req, res)
{
	console.log('Login : ' + req.body.login);

	try
	{
		pool.getConnection()
			.then(conn => {
				var query = GetQueryStringSelectPersor(req.body.login);
				conn.query(query)
					.then((rows) => {
						if (rows.length === 1)
						{
							if (rows[0].IsEnabled === 1)
							{
								var user = new Login(rows[0].ID, rows[0].PersonName, rows[0].UserIdentity, rows[0].UserPassword);
								user.otherFields = JSON.parse(rows[0].Info);

								if (user.password === req.body.pwd)
								{
									console.log("user found : " + user.email);

									// update last access date
									var query = GetQueryStringForUpdateDateLastAccess(user.id);
									conn.query(query)
										.then( async () =>
										{
											if (useSession)
											{
												console.log('create a new session');
												user.sessionId = CreateGuid();
												var remoteIp = GetRemoteIp(req);
												var querySession = GetQueryStringOpenSession(user.id, user.sessionId, remoteIp, 0);
												await conn.query(querySession);
											}
											else {
												console.log('without session');
											}

											console.log('user logged in successfully');

											var jsonUser = JSON.stringify(user);

											res.set("Connection", "close");
											res.send(jsonUser);
											conn.end();
										})
										.catch(err => {
											// update failed
											console.log('update last date access failed : ' + err);
											SendInternalServerError(res);
											conn.end();
										});
								}
								else
								{
									console.log("wrong password for user " + req.body.login);
									res.set("Connection", "close");
									res.status(401).send("wrong password");
									conn.end();
								}
							}
							else
							{
								console.log("user not enabled : " + req.body.login);
								res.set("Connection", "close");
								res.status(401).send("not enabled");
								conn.end();
							}
						}
						else
						{
							console.log("user not found : " + req.body.login);
							res.set("Connection", "close");
							res.status(404).send("not found");
							conn.end();
						}
					})
					.catch(err => {
						// select failed
						console.log('select failed : ' + err);
						SendInternalServerError(res);
						conn.end();
					})
			})
			.catch(err => {
				//not connected
				console.log('not connected : ' + err);
				SendInternalServerError(res);
			});
	}
	catch (err) {
		console.log('exception : ' + err);
		SendInternalServerError(res);
	}


});

/// <summary>
/// Verify if a user is already registered  
/// </summary>
app.get('/checklogin', urlencodedParser, function (req, res)
{
	console.log('Check login for : ' + req.query.login);

	try
	{
		pool.getConnection()
			.then(conn =>
			{
				var query = GetQueryStringSelectPersor(req.query.login);
				conn.query(query)
					.then((rows) =>
					{
						if (rows.length === 1)
						{
							console.log("user found : " + req.query.login);
							res.set("Connection", "close");
							res.send("OK");
						}
						else
						{
							console.log("user not found : " + req.query.login);
							res.set("Connection", "close");
							res.status(404).send("not found");
						}
						conn.end();							
					})
					.catch(err => {
						// select failed
						console.log('select failed : ' + err);
						SendInternalServerError(res);
						conn.end();
					})
			})
			.catch(err => {
				//not connected
				console.log('not connected : ' + err);
				SendInternalServerError(res);
			});
	}
	catch (err) {
		console.log('exception : ' + err);
		SendInternalServerError(res);
	}
});

/// <summary>
/// Verify if a user has already answered to a given quiz   
/// </summary>
app.get('/checkresult', urlencodedParser, function (req, res) {

	console.log('Check quiz result ' + req.query.name + ' for user : ' + req.query.login);

	try
	{
		pool.getConnection()
			.then(conn =>
			{
				var query = GetQueryStringForCheckResult(req.query.name, req.query.login);
				conn.query(query)
					.then((rows) => {
						if (rows.length === 1)
						{
							console.log("quiz result " + req.query.name + " found for user : " + req.query.login);

							var retakeInfo = new RetakeInfo(rows[0].NumOfRetake, rows[0].FinalMark);
							var jsonRetakeInfo = JSON.stringify(retakeInfo);

							res.set("Connection", "close");
							res.send(jsonRetakeInfo);
						}
						else
						{
							console.log("quiz result " + req.query.name + " not found for user : " + req.query.login);
							res.set("Connection", "close");
							res.status(404).send("not found");
						}
						conn.end();
					})
					.catch(err => {
						// select failed
						console.log('select failed : ' + err);
						SendInternalServerError(res);
						conn.end();
					})
			})
			.catch(err => {
				//not connected
				console.log('not connected : ' + err);
				SendInternalServerError(res);
			});
	}
	catch (err)
	{
		console.log('exception : ' + err);
		SendInternalServerError(res);
	}
});

/// <summary>
/// Get all the titles (and few other information) of the quizzes stored on database :
/// only the quizzes stored with all data, non only the results sent from users 
/// </summary>
app.get('/gettitles', urlencodedParser, function (req, res) {

	console.log('Get abstract titles');

	try {
		var quizzes = [];
		var quiz, name, title, date, id;

		pool.getConnection()
			.then(conn => {
				var query = GetQueryStringForGetTitles();
				conn.query(query)
					.then((rows) => {
						for (var i = 0; i < rows.length; i++)
						{
							id = rows[i].QZ_ID;
							name = rows[i].quizname;
							title = rows[i].Title;
							date = rows[i].DateModified;
	
							quiz = new Quiz(id, name, title, date);
							quizzes.push(quiz);
						}

						var text = JSON.stringify(quizzes);
						res.set("Connection", "close");
						res.send(text);
						conn.end();

					}).catch(err => {
						// select failed
						console.log('select failed : ' + err);
						SendInternalServerError(res);
						conn.end();
					})
			}).catch(err => {
				//not connected
				console.log('not connected : ' + err);
				SendInternalServerError(res);
			});
	}
	catch (err) {
		console.log('exception : ' + err);
		SendInternalServerError(res);
	}
});

/// <summary>
/// Get all the titles (and few other information) of the quizzes for which there are results stored
/// </summary>
app.get('/getresultsabstract', urlencodedParser, function (req, res) {

	console.log('Get abstract results');

	try
	{
		var abstracts = [];
		var abstract, title, user, date, mark;

		pool.getConnection()
			.then(conn => {
				var query = GetQueryStringForGetResultsAbstract();
				conn.query(query)
					.then((rows) => {
						for (var i = 0; i < rows.length; i++)
						{
							title = rows[i].QuizName + " - " + rows[i].QuizTitle;
							user = rows[i].UserName + " - " + rows[i].UserLogin;
							date = rows[i].DateCompleted;
							mark = rows[i].FinalMark;

							console.log('title=' + title);

							abstract = new Abstract(title, user, date, mark);
							abstracts.push(abstract);
						}

						var text = JSON.stringify(abstracts);
						res.set("Connection", "close");
						res.send(text);
						conn.end();

					}).catch(err => {
						// select failed
						console.log('select failed : ' + err);
						SendInternalServerError(res);
						conn.end();
					})
			}).catch (err => {
				//not connected
				console.log('not connected : ' + err);
				SendInternalServerError(res);
			});
	}
	catch (err)
	{
		console.log('exception : ' + err);
		SendInternalServerError(res);
	}
});

/// <summary>
/// Gets the quizzes results given some filters  
/// </summary>
app.get('/getresults', urlencodedParser, function (req, res) {

	console.log('Get quiz results');

	try {
		var result;
		var results = [];

		var searchParams = new SearchParams(
			req.query.title,
			req.query.user,
			req.query.fromDate,
			req.query.toDate,
			req.query.fromMark,
			req.query.toMark,
			req.query.top);

		pool.getConnection()
			.then(conn => {
				var query = GetQueryStringForQuizReport(searchParams);

				conn.query(query)
					.then((rows) => {

						for (var i = 0; i < rows.length; i++)
						{
							result = new Result();
							result.Header = new ResultHeader();

							result.Id = rows[i].ID;
							result.UserName = rows[i].UserName;
							result.UserLogin = rows[i].UserLogin;
							result.DateCompleted = rows[i].DateCompleted;
							result.ElapsedTime = rows[i].ElapsedTime;
							result.HighestMark = rows[i].HighestMark;
							result.NumCorrectAnswers = rows[i].RightQuestsNum;
							result.NumWrongAnswers = rows[i].WrongQuestsNum;
							result.NumNotValutated = rows[i].QuestsNum - rows[i].RightQuestsNum - rows[i].WrongQuestsNum;
							result.FinalMark = rows[i].FinalMark;

							result.Header.Name = rows[i].QuizName;
							result.Header.Title = rows[i].QuizTitle;
							result.Header.Author = rows[i].Author;
							result.Header.Argument = rows[i].Argument;
							result.Header.Duration = rows[i].TotalTime;
							result.Header.NumOfQuestions = rows[i].QuestsNum;

							results.push(result);
						}

						var text = JSON.stringify(results);
						res.set("Connection", "close");
						res.send(text);
						conn.end();

					}).catch(err => {
						// select failed
						console.log('select failed : ' + err);
						SendInternalServerError(res);
						conn.end();
					})
			}).catch(err => {
				//not connected
				console.log('not connected : ' + err);
				SendInternalServerError(res);
			});
	}
	catch (err)
	{
		console.log('exception : ' + err);
		SendInternalServerError(res);
	}

});

/// <summary>
/// Gets the questions of a quiz result, given its ID
/// </summary>
app.get('/getresultdetails', urlencodedParser, function (req, res) {

	console.log('Get quiz results details');

	try
	{
		var item;
		var items = [];

		var id = req.query.id;
		console.log('Search quiz results with ID=' + id);

		pool.getConnection()
			.then(conn => {
				var query = GetQueryStringForQuestionReport(id);

				conn.query(query)
					.then( async (rows) => {
						
						console.log('found quiz results');
						for (var i = 0; i < rows.length; i++)
						{
							item = new ResultItem();
							item.TypeOfQuest = rows[i].QuestType;
							item.Weight = rows[i].Weight;
							item.SelectedAnswers = (rows[i].QuestType === 3) ? rows[i].AdditionalTexts : rows[i].Choices;
							item.Response = rows[i].Score + " [" + rows[i].MinScore + " - " + rows[i].MaxScore + "]"; 
							item.ShortTextQuestion = rows[i].ShortTextQuestion;
							item.Valid = rows[i].Valid;

							items.push(item);
						}

						console.log('getting report..');

						var queryReport = GetQueryStringForQuizResultReport(id);
						var resultReport = await conn.query(queryReport);
						
						var report = '';
						if (resultReport.length === 1) {
							console.log('found report');
							report = resultReport[0].Report;
						}

						var text = JSON.stringify({ Answers: items, Report: report});
						res.set("Connection", "close");
						res.send(text);
						conn.end();

					}).catch(err => {
						// query select failed
						console.log('select failed : ' + err);
						SendInternalServerError(res);
						conn.end();
					})
			}).catch(err => {
				//not connected
				console.log('not connected : ' + err);
				SendInternalServerError(res);
			});
	}
	catch (err)
	{
		console.log('exception : ' + err);
		SendInternalServerError(res);
	}
});

function HandleErrorInTransaction(res, conn, msg, err) {
	console.log(msg + ' : ' + err);
	conn.query("ROLLBACK").then(() => { conn.end(); });
	SendInternalServerError(res);
}

function HandleError(res, conn, msg, err) {
	console.log(msg + ' : ' + err);
	conn.end();
	SendInternalServerError(res);
}

function SendInternalServerError(res) {
	res.set("Connection", "close");
	res.status(500).send("internal server error");
}

function CreateGuid() {
	return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
		var r = Math.random() * 16 | 0, v = c === 'x' ? r : (r & 0x3 | 0x8);
		return v.toString(16);
	});
}

function GetRemoteIp(req) {
	return (typeof req.headers['x-forwarded-for'] === 'string'
		&& req.headers['x-forwarded-for'].split(',').shift())
		|| req.connection.remoteAddress
		|| req.socket.remoteAddress
		|| req.connection.socket.remoteAddress
}







